from ovo.core.database import *
from ovo.core.storage import Storage
from ovo.core.aws import AWSSessionManager
from ovo.core.scheduler import *
from ovo.core.auth import get_username
from ovo.core.configuration import (
    load_config,
    DEFAULT_OVO_HOME,
    GLOBAL_OVO_HOME,
    get_shell_config_path,
    get_source_command,
)
from ovo.core.plugins import plugin_modules
from ovo.cli.common import console, OVONotInitializedError

# note: ovo.core.logic modules are imported at the bottom to avoid circular dependencies
from rich.panel import Panel
import os
import sys
import importlib.metadata

__version__ = importlib.metadata.version("ovo")

try:
    if sys.platform == "win32":
        console.print(
            Panel.fit(
                "[bold red]⚠️  WARNING: Windows Detected ⚠️[/bold red]\n\n"
                "[yellow]OVO workflows use Nextflow which is not supported on Windows.[/yellow]\n"
                "[yellow]Please use Linux or macOS, or run OVO in WSL2 (Windows Subsystem for Linux) to enable submitting workflows.[/yellow]",
                border_style="red",
                title="[bold red]Platform Warning[/bold red]",
            )
        )

    config = load_config(home_dir=GLOBAL_OVO_HOME or os.getenv("OVO_HOME") or DEFAULT_OVO_HOME)

    db = SqlDBEngine(db_url=config.db.url, verbose=config.db.verbose, read_only=config.props.read_only)
    db.init()

    storage = Storage(
        storage_root=config.storage.path,
        verbose=config.storage.verbose,
        aws=AWSSessionManager(assume_role=config.storage.aws.assume_role_arn, region_name=config.storage.aws.region)
        if config.storage.aws
        else None,
    )

    os.environ["NXF_HOME"] = config.nextflow_home

    schedulers: dict[str, Scheduler] = {}
    # make sure default scheduler is first in the dict
    sorted_scheduler_configs = sorted(
        config.schedulers.items(), key=lambda p: p[0] == config.default_scheduler, reverse=True
    )
    for scheduler_key, scheduler_config in sorted_scheduler_configs:
        if scheduler_config.type not in SchedulerTypes.REGISTERED_CLASSES:
            raise ValueError(
                f"Scheduler type '{scheduler_config.type}' found in config is not registered in this OVO server, make sure it was registered and imported. Supported schedulers: {SchedulerTypes.REGISTERED_CLASSES.keys()}"
            )
        init_args = {}
        if scheduler_config.aws:
            init_args["aws"] = AWSSessionManager(
                assume_role=scheduler_config.aws.assume_role_arn, region_name=scheduler_config.aws.region
            )
        SchedulerClass = SchedulerTypes.REGISTERED_CLASSES[scheduler_config.type]
        schedulers[scheduler_key] = SchedulerClass(
            name=scheduler_config.name,
            workdir=scheduler_config.workdir,
            reference_files_dir=config.reference_files_dir,
            allow_submit=True,
            submission_args=scheduler_config.submission_args,
            **init_args,
        )

    if not config.default_scheduler or config.default_scheduler not in schedulers:
        raise ValueError(f"default_scheduler '{config.default_scheduler}' not found, please check your config file.")

    if not config.local_scheduler or config.local_scheduler not in schedulers:
        raise ValueError(f"local_scheduler '{config.local_scheduler}' not found, please check your config file.")

    default_scheduler = schedulers[config.default_scheduler]
    local_scheduler = schedulers[config.local_scheduler]

    def get_scheduler(scheduler_key: str) -> Scheduler:
        if scheduler_key not in schedulers:
            raise ValueError(
                f"Scheduler '{scheduler_key}' is not registered in this OVO server, "
                f"registered schedulers: {schedulers.keys()}"
            )
        return schedulers[scheduler_key]

    # Load plugin modules to register their classes, use reload to support streamlit live reload
    for submodule_name in plugin_modules:
        submodule = importlib.import_module(submodule_name)
        importlib.reload(submodule)

    # Additional imports at the bottom to avoid circular dependencies
    from ovo.core.logic import design_logic, descriptor_logic, job_logic, project_logic, import_export_logic

except OVONotInitializedError as e:
    if ("init" in sys.argv and "home" in sys.argv) or "-h" in sys.argv or "--help" in sys.argv:
        # Ignore initialization error if we are initializing the home dir
        pass
    else:
        if os.getenv("OVO_HOME"):
            message = str(e)
        else:
            message = "\nPlease initialize OVO using [bold]ovo init home[/bold] or set [green]OVO_HOME[/green] env var."
            if shell_config_path := get_shell_config_path():
                with open(shell_config_path, "r") as f:
                    contents = f.read()
                if "OVO_HOME" in contents:
                    message = (
                        f"\n[green]OVO_HOME[/green] env variable not set.\nLooks like you have already initialized OVO and added [green]OVO_HOME[/green] "
                        f"to your {os.path.basename(shell_config_path)}, please run:\n" + get_source_command()
                    )
        console.print(Panel.fit(message, title="⚠️  OVO not initialized", border_style="red"))
        sys.exit(2)



import glob
import os
import time
import numpy as np
import argparse
from colabdesign import mk_af_model
import json


def add_cyclic_offset(self, offset_type=2):
    """add cyclic offset to connect N and C term"""

    def cyclic_offset(L):
        i = np.arange(L)
        ij = np.stack([i, i + L], -1)
        offset = i[:, None] - i[None, :]
        c_offset = np.abs(ij[:, None, :, None] - ij[None, :, None, :]).min((2, 3))
        if offset_type == 1:
            c_offset = c_offset
        elif offset_type >= 2:
            a = c_offset < np.abs(offset)
            c_offset[a] = -c_offset[a]
        if offset_type == 3:
            idx = np.abs(c_offset) > 2
            c_offset[idx] = (32 * c_offset[idx]) / abs(c_offset[idx])
        return c_offset * np.sign(offset)

    idx = self._inputs["residue_index"]
    offset = np.array(idx[:, None] - idx[None, :])
    if self.protocol == "binder":
        c_offset = cyclic_offset(self._binder_len)
        offset[self._target_len :, self._target_len :] = c_offset
    if self.protocol in ["fixbb", "partial", "hallucination"]:
        Ln = 0
        for L in self._lengths:
            offset[Ln : Ln + L, Ln : Ln + L] = cyclic_offset(L)
            Ln += L
    self._inputs["offset"] = offset


def flip_chains(pdb_str, out_path):
    """
    Parses a PDB file content string, swaps chain A with chain B so that binder is A and target is B, and writes the result to out_path.
    Assumes only chains A and B are present.

    :param pdb_str: String containing PDB file contents
    :param out_path: Path to output the modified PDB file
    """
    a_lines = []
    b_lines = []

    for line in pdb_str.splitlines():
        if line.startswith(("ATOM", "HETATM")):
            chain = line[21]
            if chain == "A":
                b_lines.append(line[:21] + "B" + line[22:])
            elif chain == "B":
                a_lines.append(line[:21] + "A" + line[22:])
            else:
                raise ValueError(f"Expected only chains A and B, found {chain}")
        elif not line.strip() or not line.startswith(("MODEL", "END")):
            raise ValueError(f"Unsupported line in ColabDesign PDB: {line}")

    with open(out_path, "w") as f:
        f.write("\n".join(a_lines) + "\n")
        f.write("\n".join(b_lines) + "\n")


def get_pdb_total_length(pdb_path):
    unique_residues = set()

    with open(pdb_path, "r") as f:
        for line in f:
            if line.startswith(("ATOM", "HETATM")):
                chain_id = line[21].strip()
                res_num = line[22:26].strip()
                ins_code = line[26].strip()  # column 27
                unique_residues.add((chain_id, res_num, ins_code))

    return len(unique_residues)


METRICS = {
    "rmsd": "target_aligned_binder_rmsd",
    "plddt": "binder_plddt",  # 0-100
    "pae": "binder_pae",  # pAE of binder chain
    "ptm": "ptm",  # predicted TM score of the whole binder-target complex, largely depending on target (0 = worst, 1 = best)
    "con": "con_loss",  # intramolecular contacts loss of each binder residue to 2 nearest neighbors (by default) within the chain, excluding immediate sequence neighbours
    "i_pae": "ipae",
    "i_ptm": "iptm",  # interaction predicted TM score (0 = worst, 1 = best)
    "i_con": "i_con_loss",  # intermolecular contacts loss of each binder residue to 1 nearest neighbor (by default) in the target chain, limited to the target hotspot if specified
}

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("input_dir", type=str)
    parser.add_argument("output_name", type=str)
    parser.add_argument("--params", required=True, type=str, help="Path to AlphaFold2 parameter data dir")
    parser.add_argument(
        "--num-recycles", type=int, default=3, help="Number of AlphaFold2 recycles (0, 1, 2, 3, ..., default 1)"
    )
    parser.add_argument(
        "--use-binder-template",
        action="store_true",
        default=False,
        help="Include binder structure as in initial guess template",
    )
    parser.add_argument(
        "--use-interface-template",
        action="store_true",
        default=False,
        help="Include target-binder as single structure in initial guess template to inform about their interface",
    )
    parser.add_argument("--blind", action="store_true", default=False, help="Do NOT use AlphaFold initial guess")
    parser.add_argument(
        "--multimer", action="store_true", default=False, help="Use AlphaFold multimer model (default = monomer)"
    )
    parser.add_argument(
        "--cyclic",
        action="store_true",
        default=False,
        help="Add cyclic offset to connect N and C term in binder (chain A)",
    )
    parser.add_argument(
        "--hotspot",
        default=None,
        help="Target hotspot positions - used only to compute contact loss metric (i_con), comma-separated",
    )
    options = parser.parse_args()

    if options.blind:
        assert not options.use_binder_template, "Cannot use both --blind and --use-binder-template"
        assert not options.use_interface_template, "Cannot use both --blind and --use-interface-template"

    if options.use_interface_template:
        assert options.use_binder_template, "--use-interface-template requires --use-binder-template"

    model = mk_af_model(
        protocol="binder",
        data_dir=options.params,
        use_multimer=options.multimer,
        model_names=["model_1_multimer_v3" if options.multimer else "model_1_ptm"],
        use_initial_guess=not options.blind,
    )
    paths = sorted(glob.glob(os.path.join(options.input_dir, "*.pdb")))
    print(f"Getting sequence lengths from {len(paths):,} PDBs")

    total_lengths = {}
    for pdb_path in paths:
        total_lengths[pdb_path] = get_pdb_total_length(pdb_path)

    if (num_distinct_lengths := len(set(total_lengths.values()))) > 1:
        print(f"Processing {len(paths)} PDBs with {num_distinct_lengths} distinct lengths.")
        print("Each length change will cause a spike in AF2 duration!")
    else:
        print(f"Processing {len(paths)} PDBs with identical lengths. This is optimal for AF2.")

    # sort paths by total sequence length
    # to avoid spikes in duration caused by changes in input length
    paths = sorted(paths, key=lambda p: total_lengths[p])

    os.makedirs(options.output_name, exist_ok=True)
    with open(options.output_name.rstrip("/") + ".jsonl", "wt") as f:
        for i, path in enumerate(paths, start=1):
            basename = os.path.basename(path).removesuffix(".pdb")
            print(f"Predicting PDB {i:,}/{len(paths):,}: {basename}")
            start_time = time.time()
            model.prep_inputs(
                path,
                # TODO add support for multiple target or binder chains
                # We can take inspiration from here: https://github.com/sokrypton/ColabDesign/blob/4127b5ab889f5b62a56644d3d1cbdd5cb313a0d0/colabdesign/rf/refolding_test.py#L87-L98
                # A comma-separated list can be passed here
                binder_chain="A",
                target_chain="B",
                rm_target=False,
                rm_binder=not options.use_binder_template,
                rm_template_ic=not options.use_interface_template,
                # Hotspots are used for connectivity loss
                # NOTE that rfdiffusion renumbers the target chain from 1 so you need to recalculate the position numbers
                hotspot=options.hotspot if options.hotspot else None,
            )
            if options.cyclic:
                add_cyclic_offset(model, offset_type=2)
            model.set_seq(mode="wildtype")
            model.set_opt(num_recycles=options.num_recycles)
            model.predict(num_models=1, verbose=False)
            metrics = {"id": basename}
            metrics.update({new_key: model.aux["log"].get(old_key) for old_key, new_key in METRICS.items()})
            metrics["binder_plddt"] *= 100
            metrics["binder_pae"] = (
                metrics["binder_pae"] * 31.0
            )  # de-normalization of https://github.com/sokrypton/ColabDesign/blob/4c0bc6d67f8f967135ecccc135a26b3bfded25e8/colabdesign/af/loss.py#L252
            metrics["ipae"] = metrics["ipae"] * 31.0
            metrics["time"] = time.time() - start_time
            metrics_str = " | ".join(f"{k} = {v:.2f}" for k, v in metrics.items() if isinstance(v, float))
            print(" Prediction done in {:.1f}s | {}".format(metrics["time"], metrics_str))
            json.dump(metrics, f)
            f.write("\n")
            f.flush()
            suffix = os.path.basename(options.output_name.rstrip("/"))
            flip_chains(model.save_pdb(), os.path.join(options.output_name, f"{basename}_{suffix}.pdb"))



from ovo import db, get_username
from ovo.core.database.models import Round


def get_or_create_project_rounds(project_id: str) -> dict[str, Round]:
    project_rounds = db.select(Round, project_id=project_id, order_by="created_date_utc")
    if not project_rounds:
        project_round = Round(project_id=project_id, name="Round 1", author=get_username())
        db.save(project_round)
        project_rounds = [project_round]
    return {r.id: r for r in project_rounds}
